/* This file is part of Mailfromd.
   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011 Sergey
   Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include "libmf.h"
#include "xalloc.h"


struct cache_list_elt {
	struct cache_list_elt *next;
	struct mf_option_cache *opt;
	size_t size;
};

static struct cache_list_elt *head, *tail;

static struct mf_option_cache *
optcache_dup(struct mf_option_cache *tab, size_t size)
{
	struct mf_option_cache *newtab = xcalloc(size, sizeof(*tab));
	size_t i;

	for (i = 0; i < size; i++) {
		newtab[i] = tab[i];
		newtab[i].name = xstrdup(newtab[i].name);
	}
	return newtab;
}

void
mf_optcache_add(struct mf_option_cache *tab, size_t size, int flags)
{
	struct cache_list_elt *elt;

	elt = xmalloc(sizeof(*elt));
	elt->next = NULL;
	if (tail)
		tail->next = elt;
	else
		head = elt;
	tail = elt;

	/* If the array is null-terminated, compute its size */
	if (flags & MF_OCF_NULL)
		for (size = 0; tab[size].name; size++)
			;

	/* Unless it is static, make a copy */
	tab = optcache_dup(tab, size);

	elt->opt = tab;
	elt->size = size;
}

static struct mf_option_cache *
find_option(char *name)
{
	struct cache_list_elt *elt;

	for (elt = head; elt; elt = elt->next) {
		size_t i;
		struct mf_option_cache *opt;
		
		for (i = 0, opt = elt->opt; i < elt->size; i++, opt++) {
			if (strcmp(opt->name, name) == 0)
				return opt;
		}
	}
	return NULL;
}

int
mf_optcache_set_option(char *name, char *value)
{
	struct mf_option_cache *p = find_option(name);
	if (!p) {
		errno = ENOENT;
		return 1;
	}
	return p->handler(name, &p->value, value);
}

void
mf_optcache_flush()
{
	struct cache_list_elt *elt;

	for (elt = head; elt; elt = elt->next) {
		size_t i;
		struct mf_option_cache *opt;
		
		for (i = 0, opt = elt->opt; i < elt->size; i++, opt++)
			if (opt->value && opt->set)
				opt->set(opt->value);
	}
}


int
mf_option_string(char *opt, void **pval, char *newval)
{
	if (*pval)
		free(*pval);
	*pval = strdup(newval);
	return 0;
}

int
mf_option_boolean(char *opt, void **pval, char *newval)
{
	int val;

	if (strcmp (newval, "yes") == 0
	    || strcmp (newval, "true") == 0
	    || strcmp (newval, "t") == 0)
		val = 1;
	else if (strcmp (newval, "no") == 0
		 || strcmp (newval, "false") == 0
		 || strcmp (newval, "nil") == 0)
		val = 0;
	else {
		char *p;
		
		val = strtoul(newval, &p, 10);
		if (*p) {
			errno = EINVAL;
			return 1;
		}
	}
	
	*pval = (void*) val;
	return 0;
}

int
mf_option_time(char *opt, void **pval, char *newval)
{
	time_t interval;
	const char *endp;
	if (parse_time_interval(newval, &interval, &endp)) {
		mu_error(_("%s: unrecognized time format (near `%s')"),
			 opt, endp);
		return 1;
	}
	if (!*pval)
		*pval = xmalloc(sizeof(time_t));
	*(time_t*) *pval = interval;
	return 0;
}

int
mf_option_time_t(char *opt, void **pval, char *newval)
{
	if (!*pval)
		*pval = xmalloc(sizeof(time_t));
	memcpy(*pval, newval, sizeof(time_t));
	return 0;
}

int
mf_option_size_t(char *opt, void **pval, char *newval)
{
	if (!*pval)
		*pval = xmalloc(sizeof(size_t));
	memcpy(*pval, newval, sizeof(size_t));
	return 0;
}
	
