/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2008, 2010, 2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "msg.h"
#include "global.h"

static size_t nmsgs = MAX_MSGS;

static struct mu_cfg_param msg_cfg_param[] = {
	{ "max-open-messages", mu_cfg_size, &nmsgs, 0, NULL,
	  N_("Maximum number of messages to open simultaneously.") },
	{ NULL }
};

static void *
alloc_msgs()
{
	return xcalloc(nmsgs, sizeof(struct mf_message));
}

static void
destroy_msgs(void *data)
{
	struct mf_message *mtab = data;
	struct mf_message *p;
	for (p = mtab; p < mtab + nmsgs; p++) {
		bi_close_message(p);
	}
	free(mtab);
}

void
drop_current_message(void *data)
{
	int i;
	struct mf_message *tab = data;
	for (i = 0; i < nmsgs; i++)
		if (tab[i].msg && tab[i].type == MF_MSG_CURRENT) {
			bi_close_message(&tab[i]);
			break;
		}
}

MF_DECLARE_DATA(MSGTAB, alloc_msgs, destroy_msgs, drop_current_message)

static int
find_slot(struct mf_message *tab)
{
	int i;
	for (i = 0; i < nmsgs; i++)
		if (tab[i].msg == NULL)
			return i;
	return -1;
}

static int
do_msg_close(void *item, void *data)
{
	bi_close_message(item);
	return 0;
}

void
bi_close_message(struct mf_message *mp)
{
	if (mp->msg) {
		mu_stream_destroy(&mp->mstr);
		mu_stream_destroy(&mp->bstr);
		if (mp->type == MF_MSG_STANDALONE)
			mu_message_destroy(&mp->msg,
					   mu_message_get_owner(mp->msg));
		mu_list_foreach(mp->msglist, do_msg_close, NULL);
		mu_list_destroy(&mp->msglist);
		free(mp->buf);
		memset(mp, 0, sizeof *mp);
	}
}

int
bi_message_register(eval_environ_t env,
		    mu_list_t msglist, mu_message_t msg, int type)
{
	struct mf_message *msgtab = MF_GET_DATA;
	int idx = find_slot(msgtab);
	if (idx >= 0) {
		struct mf_message *mp = msgtab + idx;
		mp->msg = msg;
		mp->type = type;
		mu_list_create(&mp->msglist);
		if (msglist)
			mu_list_append(msglist, mp);
	}
	return idx;
}

int
bi_get_current_message(eval_environ_t env)
{
	int i;
	struct mf_message *tab = MF_GET_DATA;
	for (i = 0; i < nmsgs; i++)
		if (tab[i].msg && tab[i].type == MF_MSG_CURRENT)
			return i;
	return -1;
}


m4_define([<DCL_MSG>],[<
	struct mf_message *mtab = MF_GET_DATA;
	struct mf_message *$1;
			
	MF_ASSERT($2 >= 0 && $2 < nmsgs,
		  mfe_range,
		  _("invalid message descriptor"));
	$1 = mtab + $2;
	MF_ASSERT($1->msg,
		  mfe_failure,
		  _("message not open"))
>])

m4_define([<DCL_BODY>],[<
	DCL_MSG($1, $2);
	if (!$1->body) {
		int rc = mu_message_get_body($1->msg, &$1->body);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_message_get_body: %s",
			  mu_strerror(rc));
	}	  
>])

m4_define([<DCL_HDR>],[<
	DCL_MSG($1, $2);
	if (!$1->hdr) {
		int rc = mu_message_get_header($1->msg, &$1->hdr);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_message_get_header: %s",
			  mu_strerror(rc));
	}	  
>])

mu_message_t
bi_message_from_descr(eval_environ_t env, int nmsg)
{
	DCL_MSG(mp, nmsg);
	return mp->msg;
}

/* void message_close(number msg) */
MF_DEFUN(message_close, VOID, NUMBER nmsg)
{
	DCL_MSG(mp, nmsg);
	bi_close_message(mp);
}
END

/* number message_size(number nmsg) */
MF_DEFUN(message_size, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t size;
	DCL_MSG(mp, nmsg);
	
	rc = mu_message_size(mp->msg, &size);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s", mu_strerror(rc));
	MF_ASSERT(size < LONG_MAX,
		  mfe_range,
		  _("message size out of representable range"));
	MF_RETURN((long)size);
}	
END

/* number message_lines(number nmsg) */
MF_DEFUN(message_lines, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t lines;
	DCL_MSG(mp, nmsg);

	rc = mu_message_lines(mp->msg, &lines);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	MF_RETURN(lines);
}
END

/* number message_body_size(number nmsg) */
MF_DEFUN(message_body_size, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t size;
	DCL_BODY(mp, nmsg);
	
	rc = mu_body_size(mp->body, &size);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s", mu_strerror(rc));
	MF_ASSERT(size < LONG_MAX,
		  mfe_range,
		  _("body size out of representable range"));
	MF_RETURN((long)size);
}
END

/* number message_body_lines(number nmsg) */
MF_DEFUN(message_body_lines, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t lines;
	DCL_BODY(mp, nmsg);

	rc = mu_body_lines(mp->body, &lines);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	MF_RETURN(lines);
}
END


/* Headers */

/* number message_header_size(number nmsg) */
MF_DEFUN(message_header_size, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t size;
	DCL_HDR(mp, nmsg);
	
	rc = mu_header_size(mp->hdr, &size);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s", mu_strerror(rc));
	MF_ASSERT(size < LONG_MAX,
		  mfe_range,
		  _("header size out of representable range"));
	MF_RETURN((long)size);
}
END

/* number message_header_lines(number nmsg) */
MF_DEFUN(message_header_lines, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t lines;
	DCL_HDR(mp, nmsg);

	rc = mu_header_lines(mp->hdr, &lines);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	MF_RETURN(lines);
}
END

/* number message_header_count(number nmsg[; string name]) */
MF_DEFUN(message_header_count, NUMBER, NUMBER nmsg,
	 OPTIONAL, STRING name)
{
	int rc;
	size_t count;
	DCL_HDR(mp, nmsg);

	rc = mu_header_get_field_count(mp->hdr, &count);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	
	if (MF_DEFINED(name)) {
		size_t n;

		for (n = 1; n <= count; n++) {
			const char *val;
			rc = mu_header_sget_value_n(mp->hdr, name, n, &val);
			if (rc == MU_ERR_NOENT)
				break;
			else if (rc)
				MF_THROW(mfe_failure, "%s", mu_strerror(rc));
		}
		count = n - 1;
	}
	MF_RETURN(count);
}
END

/* bool message_has_header(number msg, string header[, number idx]) */
MF_DEFUN(message_has_header, NUMBER, NUMBER nmsg, STRING header,
	 OPTIONAL, NUMBER idx)
{
	int rc;
	const char *val;
	DCL_HDR(mp, nmsg);

	rc = mu_header_sget_value_n(mp->hdr, header, MF_OPTVAL(idx, 1), &val);
 	MF_RETURN(rc == 0);
}
END

/* string message_find_header(number msg, string header[, number idx]) */
MF_DEFUN(message_find_header, STRING, NUMBER nmsg, STRING header,
	 OPTIONAL, NUMBER idx)
{
	int rc;
	const char *val;
	DCL_HDR(mp, nmsg);

	rc = mu_header_sget_value_n(mp->hdr, header, MF_OPTVAL(idx, 1), &val);
	if (rc == MU_ERR_NOENT)
		MF_THROW(mfe_not_found, _("header not found"));
	else if (rc)
		MF_THROW(mfe_failure, "%s", mu_strerror(rc));
	MF_RETURN_STRING(val);
}
END

/* string message_nth_header_name(number msg, number n) */
MF_DEFUN(message_nth_header_name, STRING, NUMBER nmsg, NUMBER idx)
{
	int rc;
	const char *val;
	DCL_HDR(mp, nmsg);
	
	rc = mu_header_sget_field_name(mp->hdr, idx, &val);
	if (rc == MU_ERR_NOENT)
		MF_THROW(mfe_range, "header index out of range");
	else if (rc)
		MF_THROW(mfe_failure, "%s", mu_strerror(rc));
	MF_RETURN_STRING(val);
}
END

/* string message_nth_header_value(number msg, number n) */
MF_DEFUN(message_nth_header_value, STRING, NUMBER nmsg, NUMBER idx)
{
	int rc;
	const char *val;
	DCL_HDR(mp, nmsg);
	
	rc = mu_header_sget_field_value(mp->hdr, idx, &val);
	if (rc == MU_ERR_NOENT)
		MF_THROW(mfe_range, "header index out of range");
	else if (rc)
		MF_THROW(mfe_failure, "%s", mu_strerror(rc));
	MF_RETURN_STRING(val);
}
END


/* Multipart messages */

/* bool message_is_multipart(number msg) */
MF_DEFUN(message_is_multipart, NUMBER, NUMBER nmsg)
{
	int rc, res;
	DCL_MSG(mp, nmsg);

	rc = mu_message_is_multipart(mp->msg, &res);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	MF_RETURN(res);
}
END

/* number message_count_parts(number msg) */
MF_DEFUN(message_count_parts, NUMBER, NUMBER nmsg)
{
	int rc;
	size_t count;
	DCL_MSG(mp, nmsg);
	rc = mu_message_get_num_parts(mp->msg, &count);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	MF_RETURN(count);
}
END

/* number message_get_part(number msg, number idx) */
MF_DEFUN(message_get_part, NUMBER, NUMBER nmsg, NUMBER idx)
{
	int rc;
	mu_message_t msg;
	DCL_MSG(mp, nmsg);

	rc = mu_message_get_part(mp->msg, idx, &msg);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "%s",
		  mu_strerror(rc));
	rc = bi_message_register(env, mp->msglist, msg, MF_MSG_MAILBOX);
	MF_ASSERT(rc >= 0,
		  mfe_failure,
		  _("no more message slots available"));
	MF_RETURN(rc);
}
END


/* Reading */

/* string message_rewind(number nmsg) */
MF_DEFUN(message_rewind, VOID, NUMBER nmsg)
{
	DCL_MSG(mp, nmsg);
	mu_stream_destroy(&mp->mstr);
}
END

/* string message_read_line(number nmsg) */
MF_DEFUN(message_read_line, STRING, NUMBER nmsg)
{
	int rc;
	size_t size;
	DCL_MSG(mp, nmsg);

	if (!mp->mstr) {
		rc = mu_message_get_streamref(mp->msg, &mp->mstr);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_message_get_stream: %s",
			  mu_strerror(rc));
	}
	rc = mu_stream_getline(mp->mstr, &mp->buf, &mp->bufsize, &size);
	MF_ASSERT(rc == 0,
		  mfe_io,
		  "mu_stream_getline: %s",
		  mu_strerror(rc));
	MF_ASSERT(size > 0,
		  mfe_eof,
		  "mu_stream_getline: %s",
		  _("end of input"));
	MF_RETURN_STRING(mp->buf);
}
END

/* string message_body_rewind(number nmsg) */
MF_DEFUN(message_body_rewind, VOID, NUMBER nmsg)
{
	DCL_BODY(mp, nmsg);
	mu_stream_destroy(&mp->bstr);
}
END

/* string message_read_body_line(number msg)  */
MF_DEFUN(message_read_body_line, STRING, NUMBER nmsg)
{
	int rc;
	size_t size;
	DCL_BODY(mp, nmsg);
	
	if (!mp->bstr) {
		rc = mu_body_get_streamref(mp->body, &mp->bstr);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_body_get_stream: %s",
			  mu_strerror(rc));
	}
	rc = mu_stream_getline(mp->bstr, &mp->buf, &mp->bufsize, &size);
	MF_ASSERT(rc == 0,
		  mfe_io,
		  "mu_stream_getline: %s",
		  mu_strerror(rc));
	MF_ASSERT(size > 0,
		  mfe_eof,
		  "mu_stream_getline: %s",
		  _("end of input"));
	MF_RETURN_STRING(mp->buf);
}
END

/* string message_to_stream(number fd, number msg; string mu_filter_chain) */
MF_DEFUN(message_to_stream, VOID, NUMBER fd, NUMBER nmsg, OPTIONAL,
	 STRING fltchain)
{
	int rc;
	int yes = 1;
	mu_stream_t dst;
	char *flts = MF_OPTVAL(fltchain);
	
	DCL_MSG(mp, nmsg);
	
	rc = mu_fd_stream_create(&dst, NULL, _bi_io_fd(env, fd, 1),
				 MU_STREAM_WRITE);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_fd_stream_create: %s",
		  mu_strerror(rc));
	mu_stream_ioctl(dst, MU_IOCTL_FD, MU_IOCTL_FD_SET_BORROW, &yes);
	if (flts) {
		struct mu_wordsplit ws;
		mu_stream_t flt;
		
		rc = mu_wordsplit(flts, &ws, MU_WRDSF_DEFFLAGS);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_wordsplit: %s",
			  mu_wordsplit_strerror(&ws));
		rc = mu_filter_chain_create(&flt, dst, MU_FILTER_ENCODE,
					    MU_STREAM_WRITE,
					    ws.ws_wordc, ws.ws_wordv);
		mu_wordsplit_free(&ws);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_filter_chain_create(\"%s\"): %s",
			  flts,
			  mu_strerror(rc));
		mu_stream_unref(dst);
		dst = flt;
	}

	if (!mp->mstr) {
		rc = mu_message_get_streamref(mp->msg, &mp->mstr);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_message_get_stream: %s",
			  mu_strerror(rc));
	}
	rc = mu_stream_copy(dst, mp->mstr, 0, NULL);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_stream_copy: %s",
		  mu_strerror(rc));
	mu_stream_close(dst);
	mu_stream_unref(dst);
}
END

/* number message_from_stream(number fd; string mu_filter_chain) */
MF_DEFUN(message_from_stream, NUMBER, NUMBER fd, OPTIONAL,
	 STRING fltchain)
{
	int rc;
	int yes = 1;
	mu_stream_t src;
	char *flts = MF_OPTVAL(fltchain);
	mu_message_t msg;
	
	rc = mu_fd_stream_create(&src, NULL, _bi_io_fd(env, fd, 0),
				 MU_STREAM_READ|MU_STREAM_SEEK);
				    
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_fd_stream_create: %s",
		  mu_strerror(rc));
	mu_stream_ioctl(src, MU_IOCTL_FD, MU_IOCTL_FD_SET_BORROW, &yes);
	if (flts) {
		struct mu_wordsplit ws;
		mu_stream_t flt;
		
		rc = mu_wordsplit(flts, &ws, MU_WRDSF_DEFFLAGS);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_wordsplit: %s",
			  mu_wordsplit_strerror(&ws));
		rc = mu_filter_chain_create(&flt, src, MU_FILTER_ENCODE,
					    MU_STREAM_READ,
					    ws.ws_wordc, ws.ws_wordv);
		mu_wordsplit_free(&ws);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_filter_chain_create(\"%s\"): %s",
			  flts,
			  mu_strerror(rc));
		mu_stream_unref(src);
		src = flt;
	}

	msg = MF_STREAM_TO_MESSAGE(src);

	rc = bi_message_register(env, NULL, msg, MF_MSG_STANDALONE);
	if (rc < 0) {
		mu_message_destroy(&msg, mu_message_get_owner(msg));
		MF_THROW(mfe_failure,
			 _("no more message slots available"));
	}
	MF_RETURN(rc);
}
END

/* string message_body_to_stream(number fd, number msg; string mu_filter_chain) */
MF_DEFUN(message_body_to_stream, VOID, NUMBER fd, NUMBER nmsg, OPTIONAL,
	 STRING fltchain)
{
	int rc;
	int yes = 1;
	mu_stream_t dst;
	char *flts = MF_OPTVAL(fltchain);
	
	DCL_BODY(mp, nmsg);
	
	rc = mu_fd_stream_create(&dst, NULL, _bi_io_fd(env, fd, 1),
				 MU_STREAM_WRITE);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_fd_stream_create: %s",
		  mu_strerror(rc));
	mu_stream_ioctl(dst, MU_IOCTL_FD, MU_IOCTL_FD_SET_BORROW, &yes);
	if (flts) {
		struct mu_wordsplit ws;
		mu_stream_t flt;
		
		rc = mu_wordsplit(flts, &ws, MU_WRDSF_DEFFLAGS);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_wordsplit: %s",
			  mu_wordsplit_strerror(&ws));
		rc = mu_filter_chain_create(&flt, dst, MU_FILTER_ENCODE,
					    MU_STREAM_WRITE,
					    ws.ws_wordc, ws.ws_wordv);
		mu_wordsplit_free(&ws);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_filter_chain_create(\"%s\"): %s",
			  flts,
			  mu_strerror(rc));
		mu_stream_unref(dst);
		dst = flt;
	}

	if (!mp->bstr) {
		rc = mu_body_get_streamref(mp->body, &mp->bstr);
		MF_ASSERT(rc == 0,
			  mfe_failure,
			  "mu_body_get_stream: %s",
			  mu_strerror(rc));
	}
	rc = mu_stream_copy(dst, mp->bstr, 0, NULL);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_stream_copy: %s",
		  mu_strerror(rc));
	mu_stream_close(dst);
	mu_stream_unref(dst);
}
END

MF_INIT([<
	 mf_add_runtime_params(msg_cfg_param);
	 >])
