/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2007, 2008, 2009, 2010, 2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <mailutils/mailutils.h>
#include <mflib/sieve.h>
#include "msg.h"

static void
_sieve_text_action_log(void *env_ptr,
		       mu_stream_t stream, size_t msgno,
		       mu_message_t msg,
		       const char *action, const char *fmt, va_list ap)
{
	eval_environ_t env = env_ptr;
	const char *id;
	
	mu_stream_printf(stream, "\033s<%d>", MU_LOG_NOTICE);
	id = mailfromd_msgid(env_get_context(env));
	if (id && id[0])
		/* ID includes trailing semicolon */
		mu_stream_printf(stream, "%s", id);
	mu_stream_printf(stream, "%s", action);
	if (fmt && fmt[0]) {
		mu_stream_printf(stream, "; ");
		mu_stream_vprintf(stream, fmt, ap);
	}
	mu_stream_printf(stream, "\n");
}

static void
_sieve_file_action_log(void *env_ptr,
		       mu_stream_t stream, size_t msgno,
		       mu_message_t msg,
		       const char *action, const char *fmt, va_list ap)
{
	eval_environ_t env = env_ptr;
	struct locus locus;

	_sieve_text_action_log(env_ptr, stream, msgno, msg, action, fmt, ap);

	env_get_locus(env, &locus);
	mu_stream_printf(stream, "\033f<%d>%s\033l<%lu>\033s<%d>\033O<%d>%s\n",
			 strlen(locus.file), locus.file, locus.line,
			 MU_LOG_NOTICE,
		         MU_LOGMODE_LOCUS|MU_LOGMODE_SEVERITY,
			 _("sieve called from here"));
}

static void
mach_cleanup(void *ptr)
{
	mu_sieve_machine_t mach = ptr;
	mu_sieve_machine_destroy(&mach);
}

MF_DEFUN(sieve, NUMBER, NUMBER nmsg, STRING script, OPTIONAL, NUMBER flags,
	 STRING file, NUMBER line)
{
	mu_sieve_machine_t mach;
	int sieve_debug_flags = 0;
	int rc;
	int retval = 0;
	int f = MF_OPTVAL(flags);
	mu_attribute_t attr;
	mu_message_t msg;

	rc = mu_sieve_machine_init(&mach);
	MF_ASSERT(rc == 0, mfe_failure,
		  _("failed to initialize sieve machine: %s"),
		  mu_strerror(rc));
	MF_DCL_CLEANUP(mach, mach_cleanup);
	
	if (f & MF_SIEVE_DEBUG_TRACE)
		sieve_debug_flags |= MU_SIEVE_DEBUG_TRACE;
	if (f & MF_SIEVE_DEBUG_INSTR)
		sieve_debug_flags |= MU_SIEVE_DEBUG_INSTR;

	mu_sieve_set_debug_level(mach, sieve_debug_flags);

	mu_sieve_set_data(mach, env);

	if (f & MF_SIEVE_TEXT) {
		struct locus locus;

		env_get_locus(env, &locus);
		if (f & MF_SIEVE_LOG)
			mu_sieve_set_logger(mach, _sieve_text_action_log);
		rc = mu_sieve_compile_buffer(mach, script, strlen(script),
					     MF_OPTVAL(file, locus.file),
					     MF_OPTVAL(line, locus.line));
	} else {
		if (f & MF_SIEVE_LOG)
			mu_sieve_set_logger(mach, _sieve_file_action_log);
		rc = mu_sieve_compile(mach, script);
	}

	MF_ASSERT(rc == 0,
		  mfe_failure,
		  _("compilation of Sieve script %s failed"),
		  script);

	msg = bi_message_from_descr(env, nmsg);
		
	mu_message_get_attribute(msg, &attr);
	mu_attribute_unset_deleted(attr);
	rc = mu_sieve_message(mach, msg);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  _("sieving failed: %s"),
		  mu_strerror(rc));
	retval = !(mu_attribute_is_deleted(attr) == 0);

	MF_RETURN(retval);
}
END

MF_INIT
